#include "em_device.h"
#include "em_leuart.h"
#include "sl_iostream_uart.h"
#include "sl_iostream_leuart.h"
{% if "vcom" in iostream_leuart_instances and enable_config_over_swo %}
#include "sl_cos.h"
{% endif %}
// Include instance config {# Include instance config #}
{% for instance in iostream_leuart_instances %} #include "sl_iostream_leuart_{{ instance }}_config.h"
{% endfor %}
// MACROs for generating name and IRQ handler function  {# MACROs for generating name and IRQ handler function #}
#define IOSTREAM_LEUART_CONCAT_PASTER(first, second, third)        first ##  second ## third
{% set irq_number_suffix = '_IRQn' %} {% set irq_handler_suffix = '_IRQHandler' %}
{% set rx_dma_signal_suffix = '_RXDATAV' %}
#define SL_IOSTREAM_LEUART_IRQ_NUMBER(periph_nbr)        IOSTREAM_LEUART_CONCAT_PASTER(LEUART, periph_nbr, {{ irq_number_suffix }})        {# Create IRQ number #}
#define SL_IOSTREAM_LEUART_CLOCK_REF(periph_nbr)         IOSTREAM_LEUART_CONCAT_PASTER(cmuClock_, LEUART, periph_nbr)       {# Create clock reference #}
#define SL_IOSTREAM_LEUART_IRQ_HANDLER(periph_nbr)       IOSTREAM_LEUART_CONCAT_PASTER(LEUART, periph_nbr, {{ irq_handler_suffix }})  {# Create IRQ Handler functiona name#}
#define SL_IOSTREAM_LEUART_RX_DMA_SIGNAL(periph_nbr)     IOSTREAM_LEUART_CONCAT_PASTER(dmadrvPeripheralSignal_LEUART, periph_nbr, {{ rx_dma_signal_suffix }})  {# Create DMA RX Signal name#}

#if defined(SL_CATALOG_POWER_MANAGER_PRESENT)
// EM Events
#define SLEEP_EM_EVENT_MASK      (SL_POWER_MANAGER_EVENT_TRANSITION_LEAVING_EM0)
static void events_handler(sl_power_manager_em_t from,
                           sl_power_manager_em_t to);
static sl_power_manager_em_transition_event_info_t events_info =
{
  .event_mask = SLEEP_EM_EVENT_MASK,
  .on_event = events_handler,
};
static sl_power_manager_em_transition_event_handle_t events_handle;
#endif // SL_CATALOG_POWER_MANAGER_PRESENT

{% for instance in iostream_leuart_instances %}
sl_status_t sl_iostream_leuart_init_{{ instance }}(void);
{% endfor %}

// Instance(s) handle and context variable {# Generate iostream handle and context variable(s) #}
{% for instance in iostream_leuart_instances %}static sl_iostream_uart_t sl_iostream_{{ instance }};
sl_iostream_t *sl_iostream_{{ instance }}_handle = &sl_iostream_{{ instance }}.stream;
sl_iostream_uart_t *sl_iostream_uart_{{ instance }}_handle = &sl_iostream_{{ instance }};
static sl_iostream_leuart_context_t  context_{{ instance }};
static uint8_t  rx_buffer_{{ instance }}[SL_IOSTREAM_LEUART_{{ instance | upper }}_RX_BUFFER_SIZE];
sl_iostream_instance_info_t sl_iostream_instance_{{ instance }}_info = {
  .handle = &sl_iostream_{{ instance }}.stream,
  .name = "{{ instance }}",
  .type = SL_IOSTREAM_TYPE_UART,
  .periph_id = SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL_NO,
  .init = sl_iostream_leuart_init_{{ instance }},
};
{% endfor %}

{% for instance in iostream_leuart_instances %}
sl_status_t sl_iostream_leuart_init_{{ instance }}(void)
{
  sl_status_t status;
  LEUART_Init_TypeDef init_{{ instance }} = LEUART_INIT_DEFAULT;
  init_{{ instance }}.baudrate = SL_IOSTREAM_LEUART_{{ instance | upper }}_BAUDRATE;
  init_{{ instance }}.parity = SL_IOSTREAM_LEUART_{{ instance | upper }}_PARITY;
  init_{{ instance }}.stopbits = SL_IOSTREAM_LEUART_{{ instance | upper }}_STOP_BITS;
  sl_iostream_leuart_config_t config_{{ instance }} = { {# create init struct #}
    .leuart = SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL,
    .clock = SL_IOSTREAM_LEUART_CLOCK_REF(SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL_NO),
    .tx_port = SL_IOSTREAM_LEUART_{{ instance | upper }}_TX_PORT,
    .tx_pin = SL_IOSTREAM_LEUART_{{ instance | upper }}_TX_PIN,
    .rx_port = SL_IOSTREAM_LEUART_{{ instance | upper }}_RX_PORT,
    .rx_pin = SL_IOSTREAM_LEUART_{{ instance | upper }}_RX_PIN,
#if defined(LEUART_ROUTEPEN_RXPEN)
    .tx_location = SL_IOSTREAM_LEUART_{{ instance | upper }}_TX_LOC,
    .rx_location = SL_IOSTREAM_LEUART_{{ instance | upper }}_RX_LOC,
#else
    .location = SL_IOSTREAM_LEUART_{{ instance | upper }}_ROUTE_LOC,
#endif
  };

  sl_iostream_dma_config_t dma_config_{{ instance }} = {.src = (uint8_t *)&SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL->RXDATA,
                                                        .peripheral_signal = SL_IOSTREAM_LEUART_RX_DMA_SIGNAL(SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL_NO)};


  sl_iostream_uart_config_t uart_config_{{ instance }} = {
    .dma_cfg = dma_config_{{ instance }},
    .rx_buffer = rx_buffer_{{ instance }},
    .rx_buffer_length = SL_IOSTREAM_LEUART_{{ instance | upper}}_RX_BUFFER_SIZE,
    .tx_irq_number = SL_IOSTREAM_LEUART_IRQ_NUMBER(SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL_NO),
    .rx_irq_number = SL_IOSTREAM_LEUART_IRQ_NUMBER(SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL_NO),
    .lf_to_crlf = SL_IOSTREAM_LEUART_{{ instance | upper }}_CONVERT_BY_DEFAULT_LF_TO_CRLF,
    .rx_when_sleeping = SL_IOSTREAM_LEUART_{{ instance | upper }}_RESTRICT_ENERGY_MODE_TO_ALLOW_RECEPTION
  };
  uart_config_{{ instance }}.sw_flow_control = SL_IOSTREAM_LEUART_{{ instance | upper }}_FLOW_CONTROL_TYPE == uartFlowControlSoftware;
  // Instantiate leuart instance
  status = sl_iostream_leuart_init(&sl_iostream_{{ instance }},
                                   &uart_config_{{ instance }},
                                   &init_{{ instance }},
                                   &config_{{ instance }},
                                   &context_{{ instance }});
  EFM_ASSERT(status == SL_STATUS_OK);

  {% if "vcom" in iostream_leuart_instances and enable_config_over_swo %}
  // Send VCOM config to WSTK
  // LEUART does not support HWFC, always set flowcontrol to COS_CONFIG_FLOWCONTROL_NONE
  sl_cos_config_vcom( (uint32_t) SL_IOSTREAM_LEUART_{{ instance | upper }}_BAUDRATE, COS_CONFIG_FLOWCONTROL_NONE);
  {% endif %}

  return status;
}
{% endfor %}

{# Init function #}
void sl_iostream_leuart_init_instances(void)
{
  #if defined(SL_CATALOG_POWER_MANAGER_PRESENT)
  // Enable power manager notifications
  sl_power_manager_subscribe_em_transition_event(&events_handle, &events_info);
  #endif
  // Instantiate leuart instance(s) {# Initialize leuart instance(s) #}
  {% for instance in iostream_leuart_instances %}
  sl_iostream_leuart_init_{{ instance }}();
  {% endfor %}
}

{% for instance in iostream_leuart_instances %} {# Create IRQ Handler(s) #}
// {{ instance | upper }} IRQ Handler
void SL_IOSTREAM_LEUART_IRQ_HANDLER(SL_IOSTREAM_LEUART_{{ instance | upper }}_PERIPHERAL_NO)(void)
{
  sl_iostream_leuart_irq_handler(&sl_iostream_{{ instance }});
}{% endfor %}

#if defined(SL_CATALOG_POWER_MANAGER_PRESENT)
#if !defined(SL_CATALOG_KERNEL_PRESENT)
{% for instance in iostream_leuart_instances %} {# Create IRQ Handler(s) #}
sl_power_manager_on_isr_exit_t sl_iostream_leuart_{{ instance }}_sleep_on_isr_exit(void)
{
  return sl_iostream_uart_sleep_on_isr_exit(&sl_iostream_{{ instance }});
}
#endif // SL_CATALOG_KERNEL_PRESENT
{% endfor %}
static void events_handler(sl_power_manager_em_t from,
                           sl_power_manager_em_t to)
{
  (void) from;
  if (to < SL_POWER_MANAGER_EM3){
    // Only prepare for wakeup from EM2 or less, since LEUART doesn't run in EM3
    {% for instance in iostream_leuart_instances %} 
    if (sl_iostream_uart_{{ instance }}_handle->stream.context != NULL) {
      sl_iostream_uart_prepare_for_sleep(sl_iostream_uart_{{ instance }}_handle);
    }
    {% endfor %}
  }
  
}
#endif // SL_CATALOG_POWER_MANAGER_PRESENT
